@@ -3,7 +3,6 @@ from typing import Tuple
 
 import jax
 from flax import struct
-from jax import jit
 from jax import numpy as jnp
 
 BLACK = 0
@@ -13,7 +12,11 @@ BLACK_CHAR = "@"
 WHITE_CHAR = "O"
 POINT_CHAR = "+"
 
-NSEW = jnp.array([[-1, 0], [1, 0], [0, 1], [0, -1]], dtype=jnp.int32)
+dx = jnp.int32([-1, +1, 0, 0])
+dy = jnp.int32([0, 0, -1, +1])
+
+FALSE = jnp.bool_(False)
+TRUE = jnp.bool_(True)
 
 
 @struct.dataclass
@@ -36,7 +39,7 @@ class GoState:
     adj_ren_id: jnp.ndarray = jnp.zeros((2, 19 * 19, 19 * 19), dtype=jnp.bool_)
 
     # 設置可能なマスをTrueとしたマスク
-    legal_action_mask: jnp.ndarray = jnp.zeros(19 * 19, dtype=jnp.bool_)
+    legal_action_mask: jnp.ndarray = jnp.zeros(19 * 19 + 1, dtype=jnp.bool_)
 
     # 直近8回のログ
     game_log: jnp.ndarray = jnp.full(
@@ -65,12 +68,10 @@ class GoState:
     terminated: jnp.ndarray = jnp.bool_(False)  # type:ignore
 
 
-@jit
 def observe(state: GoState, player_id, observe_all=False):
     return _get_alphazero_features(state, player_id, observe_all)
 
 
-@jit
 def _get_alphazero_features(state: GoState, player_id, observe_all):
     """
     17 x (size x size)
@@ -134,8 +135,9 @@ def _get_alphazero_features(state: GoState, player_id, observe_all):
     return jnp.vstack([log, color])
 
 
-@partial(jit, static_argnums=(0,))
-def init(size: int, rng: jax.random.KeyArray) -> Tuple[jnp.ndarray, GoState]:
+def init(
+    rng: jax.random.KeyArray, size: int = 5
+) -> Tuple[jnp.ndarray, GoState]:
     curr_player = jnp.int32(jax.random.bernoulli(rng))
     return curr_player, GoState(  # type:ignore
         size=jnp.int32(size),  # type:ignore
@@ -145,13 +147,12 @@ def init(size: int, rng: jax.random.KeyArray) -> Tuple[jnp.ndarray, GoState]:
         available_ren_id=jnp.ones((2, size * size), dtype=jnp.bool_),
         liberty=jnp.zeros((2, size * size, size * size), dtype=jnp.int32),
         adj_ren_id=jnp.zeros((2, size * size, size * size), dtype=jnp.bool_),
-        legal_action_mask=jnp.zeros(size * size, dtype=jnp.bool_),
+        legal_action_mask=jnp.ones(size * size + 1, dtype=jnp.bool_),
         game_log=jnp.full((8, size * size), 2, dtype=jnp.int32),  # type:ignore
         curr_player=curr_player,  # type:ignore
     )
 
 
-@partial(jit, static_argnums=(2,))
 def step(
     state: GoState, action: int, size: int
 ) -> Tuple[jnp.ndarray, GoState, jnp.ndarray]:
@@ -159,164 +160,53 @@ def step(
     _state, reward = _update_state_wo_legal_action(state, action, size)
 
     # add legal actions
-    _state = GoState(  # type:ignore
-        size=_state.size,
-        ren_id_board=_state.ren_id_board,
-        available_ren_id=_state.available_ren_id,
-        liberty=_state.liberty,
-        adj_ren_id=_state.adj_ren_id,
-        legal_action_mask=legal_actions(_state, size),
-        game_log=_state.game_log,
-        turn=_state.turn,
-        curr_player=_state.curr_player,
-        agehama=_state.agehama,
-        passed=_state.passed,
-        kou=_state.kou,
-        komi=_state.komi,
-        terminated=_state.terminated,
+    _state = _state.replace(  # type:ignore
+        legal_action_mask=legal_actions(_state, size)
     )
 
     # update log
     new_log = jnp.roll(_state.game_log, size * size)
     new_log = new_log.at[0].set(get_board(_state))
-    _state = GoState(  # type:ignore
-        size=_state.size,
-        ren_id_board=_state.ren_id_board,
-        available_ren_id=_state.available_ren_id,
-        liberty=_state.liberty,
-        adj_ren_id=_state.adj_ren_id,
-        legal_action_mask=_state.legal_action_mask,
-        game_log=new_log,
-        turn=_state.turn,
-        curr_player=_state.curr_player,
-        agehama=_state.agehama,
-        passed=_state.passed,
-        kou=_state.kou,
-        komi=_state.komi,
-        terminated=_state.terminated,
-    )
+    _state = _state.replace(game_log=new_log)  # type:ignore
 
     return (_state.curr_player, _state, reward)
 
 
-@partial(jit, static_argnums=(2,))
 def _update_state_wo_legal_action(
     _state: GoState, _action: int, _size: int
 ) -> Tuple[GoState, jnp.ndarray]:
     _state, _reward = jax.lax.cond(
-        _action < 0,
-        lambda: _pass_move(_state, _size),
+        (_action < _size * _size),
         lambda: _not_pass_move(_state, _action),
+        lambda: _pass_move(_state, _size),
     )
 
     # increase turn
-    _state = _increase_turn(_state)
+    _state = _state.replace(turn=_state.turn + 1)  # type: ignore
 
     # change player
-    _state = _change_player(_state)
+    _state = _state.replace(curr_player=(_state.curr_player + 1) % 2)  # type: ignore
 
     return _state, _reward
 
 
-@partial(jit, static_argnums=(1,))
 def _pass_move(_state: GoState, _size: int) -> Tuple[GoState, jnp.ndarray]:
     return jax.lax.cond(
         _state.passed,
         # 連続でパスならば終局
         lambda: (
-            _update_terminated(_state),
+            _state.replace(terminated=TRUE),  # type: ignore
             _get_reward(_state, _size),
         ),
         # 1回目のパスならばStateにパスを追加してそのまま続行
-        lambda: (_set_pass(_state, True), jnp.array([0, 0])),
+        lambda: (_state.replace(passed=True), jnp.array([0, 0])),  # type: ignore
     )
 
 
-@jit
-def _increase_turn(_state: GoState) -> GoState:
-    return GoState(  # type:ignore
-        size=_state.size,
-        ren_id_board=_state.ren_id_board,
-        available_ren_id=_state.available_ren_id,
-        liberty=_state.liberty,
-        adj_ren_id=_state.adj_ren_id,
-        legal_action_mask=_state.legal_action_mask,
-        game_log=_state.game_log,
-        turn=_state.turn + 1,
-        curr_player=_state.curr_player,
-        agehama=_state.agehama,
-        passed=_state.passed,
-        kou=_state.kou,
-        komi=_state.komi,
-        terminated=_state.terminated,
-    )
-
-
-@jit
-def _change_player(_state: GoState) -> GoState:
-    return GoState(  # type:ignore
-        size=_state.size,
-        ren_id_board=_state.ren_id_board,
-        available_ren_id=_state.available_ren_id,
-        liberty=_state.liberty,
-        adj_ren_id=_state.adj_ren_id,
-        legal_action_mask=_state.legal_action_mask,
-        game_log=_state.game_log,
-        turn=_state.turn,
-        curr_player=(_state.curr_player + 1) % 2,
-        agehama=_state.agehama,
-        passed=_state.passed,
-        kou=_state.kou,
-        komi=_state.komi,
-        terminated=_state.terminated,
-    )
-
-
-@jit
-def _set_pass(_state: GoState, _pass: bool) -> GoState:
-    return GoState(  # type:ignore
-        size=_state.size,
-        ren_id_board=_state.ren_id_board,
-        available_ren_id=_state.available_ren_id,
-        liberty=_state.liberty,
-        adj_ren_id=_state.adj_ren_id,
-        legal_action_mask=_state.legal_action_mask,
-        game_log=_state.game_log,
-        turn=_state.turn,
-        curr_player=_state.curr_player,
-        agehama=_state.agehama,
-        passed=_pass,  # type:ignore
-        kou=_state.kou,
-        komi=_state.komi,
-        terminated=_state.terminated,
-    )
-
-
-@jit
-def _update_terminated(_state: GoState) -> GoState:
-    return GoState(  # type:ignore
-        size=_state.size,
-        ren_id_board=_state.ren_id_board,
-        available_ren_id=_state.available_ren_id,
-        liberty=_state.liberty,
-        adj_ren_id=_state.adj_ren_id,
-        legal_action_mask=_state.legal_action_mask,
-        game_log=_state.game_log,
-        turn=_state.turn,
-        curr_player=jnp.int32(-1),  # type:ignore
-        agehama=_state.agehama,
-        passed=_state.passed,
-        kou=_state.kou,
-        komi=_state.komi,
-        terminated=jnp.bool_(True),  # type:ignore
-    )
-
-
-@jit
 def _not_pass_move(
     _state: GoState, _action: int
 ) -> Tuple[GoState, jnp.ndarray]:
-    state = _set_pass(_state, False)
+    state = _state.replace(passed=False)  # type: ignore
     xy = _action
     agehama_before = state.agehama[_my_color(state)]
     is_illegal = _is_illegal_move(state, xy)  # 既に他の石が置かれている or コウ
@@ -326,44 +216,9 @@ def _not_pass_move(
     state = _set_stone(state, xy)
 
     # 周囲の連を調べる
-    for nsew in NSEW:  # type:ignore
-        adj_pos = (
-            jnp.array([xy // state.size, xy % state.size], dtype=jnp.int32)
-            + nsew
-        )
-        adj_xy = adj_pos[0] * state.size + adj_pos[1]
-        state = jax.lax.cond(
-            _is_off_board(adj_pos, state.size),
-            lambda: state,  # 盤外
-            lambda: jax.lax.cond(
-                state.ren_id_board[_my_color(state), adj_xy] != -1,
-                lambda: _merge_ren(state, xy, adj_xy),
-                lambda: jax.lax.cond(
-                    state.ren_id_board[_opponent_color(state), adj_xy] != -1,
-                    lambda: _set_stone_next_to_oppo_ren(state, xy, adj_xy),
-                    lambda: GoState(  # type:ignore
-                        size=state.size,
-                        ren_id_board=state.ren_id_board,
-                        available_ren_id=state.available_ren_id,
-                        liberty=state.liberty.at[
-                            _my_color(state),
-                            state.ren_id_board[_my_color(state), xy],
-                            adj_xy,
-                        ].set(1),
-                        adj_ren_id=state.adj_ren_id,
-                        legal_action_mask=state.legal_action_mask,
-                        game_log=state.game_log,
-                        turn=state.turn,
-                        curr_player=state.curr_player,
-                        agehama=state.agehama,
-                        passed=state.passed,
-                        kou=state.kou,
-                        komi=state.komi,
-                        terminated=state.terminated,
-                    ),
-                ),
-            ),
-        )
+    state = jax.lax.fori_loop(
+        0, 4, lambda i, s: _check_around_xy(i, s, xy), state
+    )
 
     # 自殺手
     is_illegal = (
@@ -377,28 +232,14 @@ def _not_pass_move(
     ) | is_illegal
 
     # コウの確認
-    state = GoState(  # type:ignore
-        size=state.size,
-        ren_id_board=state.ren_id_board,
-        available_ren_id=state.available_ren_id,
-        liberty=state.liberty,
-        adj_ren_id=state.adj_ren_id,
-        legal_action_mask=state.legal_action_mask,
-        game_log=state.game_log,
-        turn=state.turn,
-        curr_player=state.curr_player,
-        agehama=state.agehama,
-        passed=state.passed,
-        kou=jax.lax.cond(
-            kou_occurred & state.agehama[_my_color(state)] - agehama_before
-            == 1,
-            lambda: state.kou,
-            lambda: jnp.int32(-1),
-        ),
-        komi=state.komi,
-        terminated=state.terminated,
+    kou = jax.lax.cond(
+        kou_occurred & state.agehama[_my_color(state)] - agehama_before == 1,
+        lambda: state.kou,
+        lambda: jnp.int32(-1),
     )
 
+    state = state.replace(kou=kou)
+
     return jax.lax.cond(
         is_illegal,
         lambda: _illegal_move(state),
@@ -406,7 +247,39 @@ def _not_pass_move(
     )
 
 
-@jit
+def _check_around_xy(i, state, xy):
+    adj_pos = jnp.array(
+        [xy // state.size + dx[i], xy % state.size + dy[i]], dtype=jnp.int32
+    )
+    adj_xy = adj_pos[0] * state.size + adj_pos[1]
+    is_off = _is_off_board(adj_pos, state.size)
+    is_my_ren = state.ren_id_board[_my_color(state), adj_xy] != -1
+    is_opp_ren = state.ren_id_board[_opponent_color(state), adj_xy] != -1
+    replaced_state = state.replace(
+        liberty=state.liberty.at[
+            _my_color(state),
+            state.ren_id_board[_my_color(state), xy],
+            adj_xy,
+        ].set(1)
+    )  # type:ignore
+    state = jax.lax.cond(
+        ((~is_off) & (~is_my_ren) & (~is_opp_ren)),
+        lambda: replaced_state,
+        lambda: state,
+    )
+    state = jax.lax.cond(
+        ((~is_off) & (~is_my_ren) & is_opp_ren),
+        lambda: _set_stone_next_to_oppo_ren(state, xy, adj_xy),
+        lambda: state,
+    )
+    state = jax.lax.cond(
+        ((~is_off) & is_my_ren),
+        lambda: _merge_ren(state, xy, adj_xy),
+        lambda: state,
+    )
+    return state
+
+
 def _is_illegal_move(_state: GoState, _xy):
     """
     既に石があるorコウ
@@ -422,42 +295,27 @@ def _is_illegal_move(_state: GoState, _xy):
     )
 
 
-@jit
 def _illegal_move(
     _state: GoState,
 ) -> Tuple[GoState, jnp.ndarray]:
     r: jnp.ndarray = jnp.array([1, 1])  # type:ignore
-    return _update_terminated(_state), r.at[_state.turn % 2].set(-1)
+    return _state.replace(terminated=TRUE), r.at[_state.turn % 2].set(-1)  # type: ignore
 
 
-@jit
 def _set_stone(_state: GoState, _xy: int) -> GoState:
     available_ren_id = _state.available_ren_id[_my_color(_state)]
     next_ren_id = jnp.argmax(available_ren_id)
     available_ren_id = available_ren_id.at[next_ren_id].set(False)
-    return GoState(  # type:ignore
-        size=_state.size,
+    return _state.replace(  # type:ignore
         ren_id_board=_state.ren_id_board.at[_my_color(_state), _xy].set(
             next_ren_id
         ),
         available_ren_id=_state.available_ren_id.at[_my_color(_state)].set(
             available_ren_id
         ),
-        liberty=_state.liberty,
-        adj_ren_id=_state.adj_ren_id,
-        legal_action_mask=_state.legal_action_mask,
-        game_log=_state.game_log,
-        turn=_state.turn,
-        curr_player=_state.curr_player,
-        agehama=_state.agehama,
-        passed=_state.passed,
-        kou=_state.kou,
-        komi=_state.komi,
-        terminated=_state.terminated,
     )
 
 
-@jit
 def _merge_ren(_state: GoState, _xy: int, _adj_xy: int):
     ren_id_board = _state.ren_id_board.at[_my_color(_state)].get()
 
@@ -481,14 +339,16 @@ def _merge_ren(_state: GoState, _xy: int, _adj_xy: int):
     )
     liberty = liberty.at[large_id, :].set(False)
 
-    _adj_ren_id = _state.adj_ren_id.at[_my_color(_state)].get()
-
     _oppo_adj_ren_id = jax.lax.map(
-        lambda _a: jnp.where(
-            _a[large_id], _a.at[large_id].set(False).at[small_id].set(True), _a
+        lambda row: jnp.where(
+            row[large_id],
+            row.at[large_id].set(False).at[small_id].set(True),
+            row,
         ),
-        _state.adj_ren_id.at[_opponent_color(_state)].get(),
+        _state.adj_ren_id[_opponent_color(_state)],  # (361, 361)
     )
+
+    _adj_ren_id = _state.adj_ren_id.at[_my_color(_state)].get()
     _adj_ren_id = _adj_ren_id.at[small_id].set(
         jnp.logical_or(_adj_ren_id[small_id], _adj_ren_id[large_id])
     )
@@ -497,8 +357,7 @@ def _merge_ren(_state: GoState, _xy: int, _adj_xy: int):
     return jax.lax.cond(
         new_id == adj_ren_id,
         lambda: _state,
-        lambda: GoState(  # type:ignore
-            size=_state.size,
+        lambda: _state.replace(  # type:ignore
             ren_id_board=_state.ren_id_board.at[_my_color(_state)].set(
                 ren_id_board
             ),
@@ -510,20 +369,10 @@ def _merge_ren(_state: GoState, _xy: int, _adj_xy: int):
             .set(_adj_ren_id)
             .at[_opponent_color(_state)]
             .set(_oppo_adj_ren_id),
-            legal_action_mask=_state.legal_action_mask,
-            game_log=_state.game_log,
-            turn=_state.turn,
-            curr_player=_state.curr_player,
-            agehama=_state.agehama,
-            passed=_state.passed,
-            kou=_state.kou,
-            komi=_state.komi,
-            terminated=_state.terminated,
         ),
     )
 
 
-@jit
 def _set_stone_next_to_oppo_ren(_state: GoState, _xy, _adj_xy):
     oppo_ren_id = _state.ren_id_board.at[
         _opponent_color(_state), _adj_xy
@@ -554,21 +403,9 @@ def _set_stone_next_to_oppo_ren(_state: GoState, _xy, _adj_xy):
         .set(True)
     )
 
-    state = GoState(  # type:ignore
-        size=_state.size,
-        ren_id_board=_state.ren_id_board,
-        available_ren_id=_state.available_ren_id,
+    state = _state.replace(  # type:ignore
         liberty=liberty,
         adj_ren_id=adj_ren_id,
-        legal_action_mask=_state.legal_action_mask,
-        game_log=_state.game_log,
-        turn=_state.turn,
-        curr_player=_state.curr_player,
-        agehama=_state.agehama,
-        passed=_state.passed,
-        kou=_state.kou,
-        komi=_state.komi,
-        terminated=_state.terminated,
     )
 
     return jax.lax.cond(
@@ -581,7 +418,6 @@ def _set_stone_next_to_oppo_ren(_state: GoState, _xy, _adj_xy):
     )
 
 
-@jit
 def _remove_stones(_state: GoState, _rm_ren_id, _rm_stone_xy) -> GoState:
     surrounded_stones = (
         _state.ren_id_board[_opponent_color(_state)] == _rm_ren_id
@@ -590,16 +426,16 @@ def _remove_stones(_state: GoState, _rm_ren_id, _rm_stone_xy) -> GoState:
     oppo_ren_id_board = jnp.where(
         surrounded_stones, -1, _state.ren_id_board[_opponent_color(_state)]
     )
-    liberty = jax.lax.map(
-        lambda l: jnp.where((l > 0) & surrounded_stones, 1, l),
-        _state.liberty[_my_color(_state)],
-    )
+
+    my_lib = _state.liberty[_my_color(_state)]  # (2, 361, 361) => (361, 361)
+    # surrounded_stones (361) => (my_lib > 0) & surrounded_stones (361, 361)
+    liberty = jnp.where((my_lib > 0) & surrounded_stones, 1, my_lib)
+
     available_ren_id = _state.available_ren_id.at[
         _opponent_color(_state), _rm_ren_id
     ].set(True)
 
-    return GoState(  # type:ignore
-        size=_state.size,
+    return _state.replace(  # type:ignore
         ren_id_board=_state.ren_id_board.at[_opponent_color(_state)].set(
             oppo_ren_id_board
         ),
@@ -612,30 +448,22 @@ def _remove_stones(_state: GoState, _rm_ren_id, _rm_stone_xy) -> GoState:
         .set(False)
         .at[_my_color(_state), :, _rm_ren_id]
         .set(False),
-        legal_action_mask=_state.legal_action_mask,
-        game_log=_state.game_log,
-        turn=_state.turn,
-        curr_player=_state.curr_player,
         agehama=_state.agehama.at[_my_color(_state)].add(agehama),
-        passed=_state.passed,
         kou=jnp.int32(_rm_stone_xy),  # type:ignore
-        terminated=_state.terminated,
     )
 
 
-@partial(jit, static_argnums=(1,))
 def legal_actions(state: GoState, size: int) -> jnp.ndarray:
-    return jnp.logical_not(
-        jax.lax.map(
-            lambda xy: _update_state_wo_legal_action(state, xy, size)[
-                0
-            ].terminated,
-            jnp.arange(0, size * size),
-        )
+    illegal_action = jax.lax.map(
+        lambda xy: _update_state_wo_legal_action(state, xy, size)[
+            0
+        ].terminated,
+        jnp.arange(0, size * size + 1),
     )
+    legal_action = ~illegal_action
+    return legal_action.at[size * size].set(TRUE)
 
 
-@jit
 def get_board(state: GoState) -> jnp.ndarray:
     board = jnp.full_like(state.ren_id_board[BLACK], 2)
     board = jnp.where(state.ren_id_board[BLACK] != -1, 0, board)
@@ -664,17 +492,14 @@ def _show_details(state: GoState) -> None:
     print(state.kou)
 
 
-@jit
 def _my_color(_state: GoState):
     return jnp.int32(_state.turn % 2)
 
 
-@jit
 def _opponent_color(_state: GoState):
     return jnp.int32((_state.turn + 1) % 2)
 
 
-@jit
 def _is_off_board(_pos: jnp.ndarray, size) -> bool:
     x = _pos[0]
     y = _pos[1]
@@ -684,46 +509,25 @@ def _is_off_board(_pos: jnp.ndarray, size) -> bool:
     )
 
 
-@jit
-def _kou_occurred(_state: GoState, xy: int) -> bool:
+def _kou_occurred(_state: GoState, xy: int) -> jnp.ndarray:
     size = _state.size
     x = xy // size
     y = xy % size
 
     oppo_color = _opponent_color(_state)
 
-    return jnp.logical_and(
-        jnp.logical_and(
-            jnp.logical_and(
-                jnp.logical_or(
-                    x - 1 < 0,
-                    _state.ren_id_board[oppo_color][_to_xy(x - 1, y, size)]
-                    != -1,
-                ),
-                jnp.logical_or(
-                    x + 1 >= size,
-                    _state.ren_id_board[oppo_color][_to_xy(x + 1, y, size)]
-                    != -1,
-                ),
-            ),
-            jnp.logical_or(
-                y - 1 < 0,
-                _state.ren_id_board[oppo_color][_to_xy(x, y - 1, size)] != -1,
-            ),
-        ),
-        jnp.logical_or(
-            y + 1 >= size,
-            _state.ren_id_board[oppo_color][_to_xy(x, y + 1, size)] != -1,
-        ),
-    )
+    to_xy_batch = jax.vmap(partial(_to_xy, size=size))
+    oob = jnp.bool_([x - 1 < 0, x + 1 >= size, y - 1 < 0, y + 1 >= size])
+    xs = x + dx
+    ys = y + dy
+    is_occupied = _state.ren_id_board[oppo_color][to_xy_batch(xs, ys)] != -1
+    return (oob | is_occupied).all()
 
 
-@jit
 def _to_xy(x, y, size) -> int:
     return x * size + y
 
 
-@partial(jit, static_argnums=(1,))
 def _get_reward(_state: GoState, _size: int) -> jnp.ndarray:
     b = _count_ji(_state, BLACK, _size) - _state.agehama[WHITE] - _state.komi
     w = _count_ji(_state, WHITE, _size) - _state.agehama[BLACK]
@@ -734,7 +538,6 @@ def _get_reward(_state: GoState, _size: int) -> jnp.ndarray:
     return r
 
 
-@partial(jit, static_argnums=(2,))
 def _count_ji(_state: GoState, _color, _size):
     board = get_board(_state)
     return jnp.count_nonzero(_get_ji(board, _color, _size))
@@ -749,7 +552,6 @@ class JI:
     color: jnp.ndarray
 
 
-@partial(jit, static_argnums=(2,))
 def _get_ji(_board: jnp.ndarray, color: int, size: int):
     BOARD_WIDTH = size
     # 1. boardの一番外側に1周分追加
@@ -790,7 +592,6 @@ def _get_ji(_board: jnp.ndarray, color: int, size: int):
     return board[1 : BOARD_WIDTH + 1, 1 : BOARD_WIDTH + 1] == POINT
 
 
-@jit
 def _count_ji_loop(_ji: JI) -> JI:
     size = _ji.size
     board = _ji.board
