@@ -99,6 +99,7 @@ def step(state: ShogiState, action: int) -> Tuple[ShogiState, int, bool]:
         return s, 0, False
 
 
+# turnから報酬計算
 def _turn_to_reward(turn: int) -> int:
     if turn == 0:
         return 1
@@ -106,6 +107,7 @@ def _turn_to_reward(turn: int) -> int:
         return -1
 
 
+# 初期盤面生成
 def _make_init_board() -> np.ndarray:
     array = np.zeros((29, 81), dtype=np.int32)
     for i in range(81):
@@ -164,6 +166,7 @@ def _make_init_board() -> np.ndarray:
     return array
 
 
+# 盤面の情報をStateに変換
 def _make_board(bs: np.ndarray) -> ShogiState:
     board = np.zeros((29, 81), dtype=np.int32)
     for i in range(81):
@@ -174,21 +177,14 @@ def _make_board(bs: np.ndarray) -> ShogiState:
 
 def _pawn_move(turn: int) -> np.ndarray:
     array = np.zeros((9, 9), dtype=np.int32)
-    if turn == 0:
-        array[4][3] = 1
-    else:
-        array[4][5] = 1
+    array[4][4 - _turn_to_reward(turn)] = 1
     return array
 
 
 def _knight_move(turn: int) -> np.ndarray:
     array = np.zeros((9, 9), dtype=np.int32)
-    if turn == 0:
-        array[3][2] = 1
-        array[5][2] = 1
-    else:
-        array[3][6] = 1
-        array[5][6] = 1
+    array[3][4 - 2 * _turn_to_reward(turn)] = 1
+    array[5][4 - 2 * _turn_to_reward(turn)] = 1
     return array
 
 
@@ -223,6 +219,7 @@ def _king_move() -> np.ndarray:
     return array
 
 
+# 端判定
 def _is_side(point: int) -> Tuple[bool, bool, bool, bool]:
     is_up = point % 9 == 0
     is_down = point % 9 == 8
@@ -243,6 +240,7 @@ def _point_to_location(point: int) -> Tuple[int, int]:
     return point // 9, point % 9
 
 
+# 端にいる駒の動けない地点へのフラグを折る
 def _cut_outside(array: np.ndarray, point: int) -> np.ndarray:
     new_array = array
     u, d, l, r = _is_side(point)
@@ -263,6 +261,7 @@ def _cut_outside(array: np.ndarray, point: int) -> np.ndarray:
     return new_array
 
 
+# 駒種と位置から到達できる地点をすべて生成
 def _action_board(array: np.ndarray, point: int) -> np.ndarray:
     new_array = array
     y, t = _point_to_location(point)
@@ -270,6 +269,7 @@ def _action_board(array: np.ndarray, point: int) -> np.ndarray:
     return np.roll(new_array, (y - 4, t - 4), axis=(0, 1)).reshape(81)
 
 
+# 大駒・香車以外の位置ごとの移動できる地点を記録
 POINT_MOVES = np.zeros((81, 29, 81), dtype=np.int32)
 for i in range(81):
     POINT_MOVES[i][1] = _action_board(_pawn_move(0), i)
@@ -297,12 +297,12 @@ def _dlshogi_action(direction: int, to: int) -> int:
     return direction * 81 + to
 
 
-# from, toが同じ列にあるかどうか
+# from, toが同じ縦列にあるかどうか
 def _is_same_column(_from: int, to: int) -> bool:
     return _from // 9 == to // 9
 
 
-# from, toが同じ行にあるかどうか
+# from, toが同じ横列にあるかどうか
 def _is_same_row(_from: int, to: int) -> bool:
     return _from % 9 == to % 9
 
@@ -317,7 +317,8 @@ def _is_same_declining(_from: int, to: int) -> bool:
     return _from // 9 + _from % 9 == to // 9 + to % 9
 
 
-def _is_same_line(from_: int, to: int, direction: int):
+# from, からdirectionの方向に行ったときにtoに到達できるか
+def _is_same_line(from_: int, to: int, direction: int) -> bool:
     dir = direction % 10
     if dir == 0 or dir == 5:
         return _is_same_column(from_, to)
@@ -331,33 +332,16 @@ def _is_same_line(from_: int, to: int, direction: int):
         return False
 
 
-def _dis_direction_array(from_: int, turn: int, direction: int):
+# fromからある方向に移動させたときの位置と距離の関係
+# それ以上その方向に動かせないなら全部0
+def _dis_direction_array(from_: int, turn: int, direction: int) -> np.ndarray:
     array = np.zeros(81, dtype=np.int32)
     dif = _direction_to_dif(direction, turn)
-    to = from_ + dif
-    if _is_in_board(to) and _is_same_line(from_, to, direction):
-        array[to] = 1
-    to += dif
-    if _is_in_board(to) and _is_same_line(from_, to, direction):
-        array[to] = 2
-    to += dif
-    if _is_in_board(to) and _is_same_line(from_, to, direction):
-        array[to] = 3
-    to += dif
-    if _is_in_board(to) and _is_same_line(from_, to, direction):
-        array[to] = 4
-    to += dif
-    if _is_in_board(to) and _is_same_line(from_, to, direction):
-        array[to] = 5
-    to += dif
-    if _is_in_board(to) and _is_same_line(from_, to, direction):
-        array[to] = 6
-    to += dif
-    if _is_in_board(to) and _is_same_line(from_, to, direction):
-        array[to] = 7
-    to += dif
-    if _is_in_board(to) and _is_same_line(from_, to, direction):
-        array[to] = 8
+    for i in range(8):
+        dis = 1 + i
+        to = from_ + dif * dis
+        if _is_in_board(to) and _is_same_line(from_, to, direction):
+            array[to] = dis
     return array
 
 
@@ -406,7 +390,7 @@ def _hand_to_direction(piece: int) -> int:
         return 12 + piece
 
 
-# AnimalShogiActionをdlshogiのint型actionに変換
+# ShogiActionをdlshogiのint型actionに変換
 def _action_to_dlaction(action: ShogiAction, turn: int) -> int:
     if action.is_drop:
         return _dlshogi_action(_hand_to_direction(action.piece), action.to)
@@ -462,14 +446,11 @@ def _direction_to_from(
     dif = _direction_to_dif(direction, state.turn)
     f = to
     _from = -1
-    for i in range(9):
+    for i in range(8):
         f -= dif
         if _is_in_board(f) and _from == -1 and _piece_type(state, f) != 0:
             _from = f
-    if direction >= 10:
-        return _from, True
-    else:
-        return _from, False
+    return _from, direction >= 10
 
 
 def _direction_to_hand(direction: int) -> int:
@@ -481,22 +462,35 @@ def _direction_to_hand(direction: int) -> int:
         return direction - 12
 
 
-def _piece_type(state: ShogiState, point: int):
-    return state.board[:, point].argmax()
+def _piece_type(state: ShogiState, point: int) -> int:
+    return int(state.board[:, point].argmax())
+
+
+# dlshogiのactionの情報をShogiActionの情報に変換
+def _dlshogi_move_action(
+    direction: int, to: int, state: ShogiState
+) -> ShogiAction:
+    _from, is_promote = _direction_to_from(direction, to, state)
+    piece = _piece_type(state, _from)
+    captured = _piece_type(state, to)
+    return ShogiAction(False, piece, to, _from, captured, is_promote)
+
+
+def _dlshogi_drop_action(
+    direction: int, to: int, state: ShogiState
+) -> ShogiAction:
+    piece = _direction_to_hand(direction)
+    return ShogiAction(True, piece, to)
 
 
 def _dlaction_to_action(action: int, state: ShogiState) -> ShogiAction:
     direction, to = _separate_dlaction(action)
     if direction <= 19:
-        # 駒の移動
-        _from, is_promote = _direction_to_from(direction, to, state)
-        piece = _piece_type(state, _from)
-        captured = _piece_type(state, to)
-        return ShogiAction(False, piece, to, _from, captured, is_promote)
+        # 移動のaction
+        return _dlshogi_move_action(direction, to, state)
     else:
         # 駒打ち
-        piece = _direction_to_hand(direction)
-        return ShogiAction(True, piece, to)
+        return _dlshogi_drop_action(direction, to, state)
 
 
 # 手番側でない色を返す
@@ -512,7 +506,7 @@ def _is_in_board(point: int) -> bool:
 # 相手の駒を同じ種類の自分の駒に変換する
 def _convert_piece(piece: int) -> int:
     if piece == 0:
-        return -1
+        return 0
     p = (piece + 14) % 28
     if p == 0:
         return 28
@@ -537,7 +531,8 @@ def _piece_to_hand(piece: int) -> int:
 def _owner(piece: int) -> int:
     if piece == 0:
         return 2
-    return (piece - 1) // 14
+    else:
+        return (piece - 1) // 14
 
 
 # 盤面のどこに何の駒があるかをnp.arrayに移したもの
@@ -581,168 +576,82 @@ def _drop(state: ShogiState, action: ShogiAction) -> ShogiState:
     return s
 
 
+# 方向ごとの大ごまの動きのなかで一番奥の地点を返す
+def _inner_point(bs_one: np.ndarray, from_: int, direction: int) -> int:
+    dir_array = _dis_direction_array(from_, 0, direction)
+    # 途中で駒にぶつからない場合
+    if np.all(dir_array * bs_one == 0):
+        return from_ + _direction_to_dif(direction, 0) * np.max(dir_array)
+    else:
+        return _nearest_position(from_, direction, bs_one)
+
+
+# fromからtoまでの地点をdifごとに1に置き換える
+# 最大8回
+# fromは0、toは1に
+def _change_between(from_: int, to: int, dif: int) -> np.ndarray:
+    array = np.zeros(81, dtype=np.int32)
+    point = from_
+    flag = from_ != to
+    for i in range(8):
+        point += dif
+        if flag and _is_in_board(point):
+            array[point] = 1
+        if point == to:
+            flag = False
+    return array
+
+
+# 香車の動き
 def _lance_move(bs: np.ndarray, from_: int, turn: int) -> np.ndarray:
-    to = np.zeros(81, dtype=np.int32)
-    u_array = _dis_direction_array(from_, turn, 0)
     bs_one = np.where(bs == 0, 0, 1)
-    if np.all(u_array * bs_one == 0):
-        if np.all(u_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(u_array)
-    else:
-        max_dis = np.min(u_array[np.nonzero(u_array * bs_one)])
-    if turn == 0:
-        u_point = from_ - max_dis
-        np.put(to, np.arange(u_point, from_, 1), 1)
-    else:
-        d_point = from_ + max_dis
-        np.put(to, np.arange(d_point, from_, -1), 1)
-    return to
+    direction = 5 * turn
+    point = _inner_point(bs_one, from_, direction)
+    return _change_between(from_, point, _direction_to_dif(0, turn))
 
 
+# 角の動き
 def _bishop_move(bs: np.ndarray, from_: int) -> np.ndarray:
     to = np.zeros(81, dtype=np.int32)
-    ur_array = _dis_direction_array(from_, 0, 2)
-    ul_array = _dis_direction_array(from_, 0, 1)
-    dr_array = _dis_direction_array(from_, 0, 7)
-    dl_array = _dis_direction_array(from_, 0, 6)
     bs_one = np.where(bs == 0, 0, 1)
-    if np.all(bs_one * ur_array == 0):
-        if np.all(ur_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(ur_array)
-    else:
-        max_dis = np.min(ur_array[np.nonzero(ur_array * bs_one)])
-    ur_point = from_ - 10 * max_dis
-    np.put(to, np.arange(ur_point, from_, 10), 1)
-    if np.all(bs_one * ul_array == 0):
-        if np.all(ul_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(ul_array)
-    else:
-        max_dis = np.min(ul_array[np.nonzero(ul_array * bs_one)])
-    ul_point = from_ + 8 * max_dis
-    np.put(to, np.arange(ul_point, from_, -8), 1)
-    if np.all(bs_one * dr_array == 0):
-        if np.all(dr_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(dr_array)
-    else:
-        max_dis = np.min(dr_array[np.nonzero(dr_array * bs_one)])
-    dr_point = from_ - 8 * max_dis
-    np.put(to, np.arange(dr_point, from_, 8), 1)
-    if np.all(bs_one * dl_array == 0):
-        if np.all(dl_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(dl_array)
-    else:
-        max_dis = np.min(dl_array[np.nonzero(dl_array * bs_one)])
-    dl_point = from_ + 10 * max_dis
-    np.put(to, np.arange(dl_point, from_, -10), 1)
+    ur_point = _inner_point(bs_one, from_, 2)
+    to += _change_between(from_, ur_point, -10)
+    ul_point = _inner_point(bs_one, from_, 1)
+    to += _change_between(from_, ul_point, 8)
+    dr_point = _inner_point(bs_one, from_, 7)
+    to += _change_between(from_, dr_point, -8)
+    dl_point = _inner_point(bs_one, from_, 6)
+    to += _change_between(from_, dl_point, 10)
     return to
-    # for i in range(8):
-    #    ur = point - 10 * (1 + i)
-    #    ul = point + 8 * (1 + i)
-    #    dr = point - 8 * (1 + i)
-    #    dl = point + 10 * (1 + i)
-    #    if _is_in_board(ur) and _is_same_rising(point, ur) and ur_flag:
-    #        array[ur] = 1
-    #    if _is_in_board(ul) and _is_same_declining(point, ul) and ul_flag:
-    #        array[ul] = 1
-    #    if _is_in_board(dr) and _is_same_declining(point, dr) and dr_flag:
-    #        array[dr] = 1
-    #    if _is_in_board(dl) and _is_same_rising(point, dl) and dl_flag:
-    #        array[dl] = 1
-    #    if not _is_in_board(ur) or bs[ur] != 0:
-    #        ur_flag = False
-    #   if not _is_in_board(ul) or bs[ul] != 0:
-    #        ul_flag = False
-    #    if not _is_in_board(dr) or bs[dr] != 0:
-    #        dr_flag = False
-    #    if not _is_in_board(dl) or bs[dl] != 0:
-    #        dl_flag = False
-    # return array
 
 
+# 飛車の動き
 def _rook_move(bs: np.ndarray, from_: int) -> np.ndarray:
     to = np.zeros(81, dtype=np.int32)
-    u_array = _dis_direction_array(from_, 0, 0)
-    d_array = _dis_direction_array(from_, 0, 5)
-    r_array = _dis_direction_array(from_, 0, 4)
-    l_array = _dis_direction_array(from_, 0, 3)
     bs_one = np.where(bs == 0, 0, 1)
-    if np.all(bs_one * u_array == 0):
-        if np.all(u_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(u_array)
-    else:
-        max_dis = np.min(u_array[np.nonzero(u_array * bs_one)])
-    ur_point = from_ - max_dis
-    np.put(to, np.arange(ur_point, from_, 1), 1)
-    if np.all(bs_one * d_array == 0):
-        if np.all(d_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(d_array)
-    else:
-        max_dis = np.min(d_array[np.nonzero(d_array * bs_one)])
-    ul_point = from_ + max_dis
-    np.put(to, np.arange(ul_point, from_, -1), 1)
-    if np.all(bs_one * r_array == 0):
-        if np.all(r_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(r_array)
-    else:
-        max_dis = np.min(r_array[np.nonzero(r_array * bs_one)])
-    r_point = from_ - 9 * max_dis
-    np.put(to, np.arange(r_point, from_, 9), 1)
-    if np.all(bs_one * l_array == 0):
-        if np.all(l_array == 0):
-            max_dis = 0
-        else:
-            max_dis = np.max(l_array)
-    else:
-        max_dis = np.min(l_array[np.nonzero(l_array * bs_one)])
-    l_point = from_ + 9 * max_dis
-    np.put(to, np.arange(l_point, from_, -9), 1)
+    u_point = _inner_point(bs_one, from_, 0)
+    to += _change_between(from_, u_point, -1)
+    d_point = _inner_point(bs_one, from_, 5)
+    to += _change_between(from_, d_point, 1)
+    r_point = _inner_point(bs_one, from_, 4)
+    to += _change_between(from_, r_point, -9)
+    l_point = _inner_point(bs_one, from_, 3)
+    to += _change_between(from_, l_point, 9)
     return to
 
 
+# 馬の動き
 def _horse_move(bs: np.ndarray, from_: int) -> np.ndarray:
-    to = _bishop_move(bs, from_)
-    u, d, l, r = _is_side(from_)
-    if not u:
-        to[from_ - 1] = 1
-    if not d:
-        to[from_ + 1] = 1
-    if not r:
-        to[from_ - 9] = 1
-    if not l:
-        to[from_ + 9] = 1
-    return to
+    # 角の動き＋玉の動き
+    to = _bishop_move(bs, from_) + POINT_MOVES[from_][8]
+    return np.where(to == 0, 0, 1)
 
 
+# 龍の動き
 def _dragon_move(bs: np.ndarray, from_: int) -> np.ndarray:
-    to = _rook_move(bs, from_)
-    u, d, l, r = _is_side(from_)
-    if not u:
-        if not r:
-            to[from_ - 10] = 1
-        if not l:
-            to[from_ + 8] = 1
-    if not d:
-        if not r:
-            to[from_ - 8] = 1
-        if not l:
-            to[from_ + 10] = 1
-    return to
+    # 飛車の動き＋玉の動き
+    to = _rook_move(bs, from_) + POINT_MOVES[from_][8]
+    return np.where(to == 0, 0, 1)
 
 
 # 駒種と位置から到達できる場所を返す
@@ -781,28 +690,37 @@ def _is_enemy_zone(turn: int, point: int) -> bool:
         return point % 9 >= 6
 
 
+# 成れるかどうか
 def _can_promote(piece: int, _from: int, to: int) -> bool:
     # pieceが飛車以下でないと成れない
     if piece % 14 > 6 or piece % 14 == 0:
         return False
     # _fromとtoのどちらかが敵陣であれば成れる
-    return _is_enemy_zone(_owner(piece), _from) or _is_enemy_zone(
-        _owner(piece), to
-    )
+    else:
+        return _is_enemy_zone(_owner(piece), _from) or _is_enemy_zone(
+            _owner(piece), to
+        )
+
+
+def _create_one_piece_actions(
+    piece: int, _from: int, to: int, actions: np.ndarray
+) -> np.ndarray:
+    new_actions = actions
+    normal_dir = _point_to_direction(_from, to, False, _owner(piece))
+    normal_act = _dlshogi_action(normal_dir, to)
+    new_actions[normal_act] = 1
+    pro_dir = _point_to_direction(_from, to, True, _owner(piece))
+    pro_act = _dlshogi_action(pro_dir, to)
+    if _can_promote(piece, _from, to):
+        new_actions[pro_act] = 1
+    return new_actions
 
 
 def _create_actions(piece: int, _from: int, to: np.ndarray) -> np.ndarray:
     actions = np.zeros(2754, dtype=np.int32)
     for i in range(81):
-        if to[i] == 0:
-            continue
-        if _can_promote(piece, _from, i):
-            pro_dir = _point_to_direction(_from, i, True, _owner(piece))
-            pro_act = _dlshogi_action(pro_dir, i)
-            actions[pro_act] = 1
-        normal_dir = _point_to_direction(_from, i, False, _owner(piece))
-        normal_act = _dlshogi_action(normal_dir, i)
-        actions[normal_act] = 1
+        if to[i] != 0:
+            actions = _create_one_piece_actions(piece, _from, i, actions)
     return actions
 
 
@@ -860,8 +778,6 @@ def _init_legal_actions(state: ShogiState) -> ShogiState:
     # 移動の追加
     for i in range(81):
         piece = bs[i]
-        if piece == 0:
-            continue
         if piece <= 14:
             s.legal_actions_black = _add_move_actions(
                 piece, i, s.legal_actions_black
@@ -883,6 +799,14 @@ def _init_legal_actions(state: ShogiState) -> ShogiState:
     return s
 
 
+# 成駒を成る前の駒に変更
+def _degeneration_piece(piece: int) -> int:
+    if piece % 14 >= 9 or piece == 14 or piece == 28:
+        return piece - 8
+    else:
+        return piece
+
+
 # 駒の移動によるlegal_actionsの更新
 def _update_legal_move_actions(
     state: ShogiState, action: ShogiAction
@@ -906,15 +830,14 @@ def _update_legal_move_actions(
     new_player_actions = _add_move_actions(
         new_piece, action.to, new_player_actions
     )
+    # 取った駒を自分の持ち駒に変換
+    # 取っていない場合は0
+    captured = _degeneration_piece(_convert_piece(action.captured))
     # 駒が取られた場合、相手の取られた駒によってできていたactionのフラグを折る
     if action.captured != 0:
         new_enemy_actions = _filter_move_actions(
             action.captured, action.to, new_enemy_actions
         )
-        captured = _convert_piece(action.captured)
-        # 成駒の場合成る前の駒に変換
-        if captured % 14 == 0 or captured % 14 >= 9:
-            captured -= 8
         new_player_actions = _add_drop_actions(captured, new_player_actions)
     if s.turn == 0:
         s.legal_actions_black = new_player_actions
@@ -970,11 +893,18 @@ def _filter_occupied_drop_actions(
     for i in range(81):
         if owner[i] == 2:
             continue
-        np.put(
-            new_array,
-            np.arange(81 * (20 + 7 * turn) + i, 81 * (27 + 7 * turn) + i, 81),
-            0,
-        )
+        if turn == 0:
+            np.put(
+                new_array,
+                np.arange(81 * 20 + i, 81 * 27 + i, 81),
+                0,
+            )
+        else:
+            np.put(
+                new_array,
+                np.arange(81 * 27 + i, 81 * 34 + i, 81),
+                0,
+            )
     return new_array
 
 
@@ -1005,29 +935,35 @@ def _legal_actions(state: ShogiState) -> np.ndarray:
 # 王手判定
 # 密接・遠隔の王手で分ける
 def _is_check(state: ShogiState) -> Tuple[int, np.ndarray, int, np.ndarray]:
-    check_near = 0
-    check_far = 0
-    checking_point_near = np.zeros(81, dtype=np.int32)
-    checking_point_far = np.zeros(81, dtype=np.int32)
-    # そもそも王がいない場合はFalse
+    # そもそも王がいない場合
     if np.all(state.board[8 + 14 * state.turn] == 0):
-        return check_near, checking_point_near, check_far, checking_point_far
+        return 0, np.zeros(81, dtype=np.int32), 0, np.zeros(81, dtype=np.int32)
+    else:
+        return _is_check_(state)
+
+
+# 玉がいる前提
+def _is_check_(state: ShogiState) -> Tuple[int, np.ndarray, int, np.ndarray]:
+    check = 0
+    checking_point = np.zeros((2, 81), dtype=np.int32)
     king_point = int(state.board[8 + 14 * state.turn, :].argmax())
     near_king = _small_piece_moves(8 + 14 * state.turn, king_point)
     bs = _board_status(state)
     for i in range(81):
         piece = bs[i]
-        if _owner(piece) != _another_color(state):
-            continue
-        if _piece_moves(bs, piece, i)[king_point] == 1:
+        if (
+            _owner(piece) == _another_color(state)
+            and _piece_moves(bs, piece, i)[king_point] == 1
+        ):
             # 桂馬の王手も密接としてカウント
             if near_king[i] == 1 or piece % 14 == 3:
-                check_near += 1
-                checking_point_near[i] = 1
+                check += 10
+                checking_point[0][i] = 1
             else:
-                check_far += 1
-                checking_point_far[i] = 1
-    return check_near, checking_point_near, check_far, checking_point_far
+                # 遠隔の王手は9以上ありえない
+                check += 1
+                checking_point[1][i] = 1
+    return check // 10, checking_point[0], check % 10, checking_point[1]
 
 
 # 二歩判定
@@ -1048,27 +984,30 @@ def _is_double_pawn(state: ShogiState) -> bool:
 def _is_stuck(state: ShogiState) -> bool:
     is_stuck = False
     bs = _board_status(state)
-    # 先手歩
-    if np.any(np.where(bs == 1)[0] % 9 == 0):
+    line1 = bs[0::9]
+    if np.any(line1 == 1):
         is_stuck = True
-    # 先手香車
-    if np.any(np.where(bs == 2)[0] % 9 == 0):
+    if np.any(line1 == 2):
         is_stuck = True
-    # 先手桂馬
-    if np.any(np.where(bs == 3)[0] % 9 <= 1):
+    if np.any(line1 == 3):
         is_stuck = True
-    # 後手歩
-    if np.any(np.where(bs == 15)[0] % 9 == 8):
+    line2 = bs[1::9]
+    if np.any(line2 == 3):
         is_stuck = True
-    # 後手香車
-    if np.any(np.where(bs == 16)[0] % 9 == 8):
+    line8 = bs[7::9]
+    if np.any(line8 == 17):
         is_stuck = True
-    # 後手桂馬
-    if np.any(np.where(bs == 17)[0] % 9 >= 7):
+    line9 = bs[8::9]
+    if np.any(line9 == 15):
+        is_stuck = True
+    if np.any(line9 == 16):
+        is_stuck = True
+    if np.any(line9 == 17):
         is_stuck = True
     return is_stuck
 
 
+# pinされている方向から縦、横、右斜めうえ、右斜め下の四方向に分類
 def _direction_to_pin(direction: int) -> int:
     if direction == 0 or direction == 5:
         return 1
@@ -1081,6 +1020,25 @@ def _direction_to_pin(direction: int) -> int:
     return 0
 
 
+# それぞれの方向について、1番fromに近い駒の位置を返す
+def _nearest_position(from_: int, direction: int, bs_one: np.ndarray) -> int:
+    flag1 = True
+    point1 = -1
+    dif = _direction_to_dif(direction, 0)
+    for i in range(8):
+        point = from_ + (1 + i) * dif
+        if (
+            flag1
+            and _is_in_board(point)
+            and _is_same_line(from_, point, direction)
+            and bs_one[point] == 1
+        ):
+            point1 = point
+            flag1 = False
+    return point1
+
+
+# pinされている駒の位置と方向を記録(1方向のみ)
 def _direction_pin(
     bs: np.ndarray,
     turn: int,
@@ -1089,40 +1047,36 @@ def _direction_pin(
     array: np.ndarray,
 ) -> np.ndarray:
     new_array = array
-    dir_array = _dis_direction_array(king_point, 0, direction)
     e_turn = (turn + 1) % 2
     bs_one = np.where(bs == 0, 0, 1)
-    dir_one_array = dir_array * bs_one
-    if np.count_nonzero(dir_one_array) <= 1:
+    # 玉に一番近い駒の位置
+    point1 = _nearest_position(king_point, direction, bs_one)
+    # 2番目に近い駒の位置
+    point2 = _nearest_position(point1, direction, bs_one)
+    if point1 == -1 or point2 == -1:
         return new_array
-    dif = _direction_to_dif(direction, 0)
-    dis1 = np.partition(dir_one_array[dir_one_array.nonzero()].flatten(), 0)[0]
-    dis2 = np.partition(dir_one_array[dir_one_array.nonzero()].flatten(), 1)[1]
-    point1 = king_point + dis1 * dif
-    point2 = king_point + dis2 * dif
     piece1 = bs[point1]
     piece2 = bs[point2]
     if _owner(piece1) != turn:
         return new_array
     flag = False
-    if direction == 0:
-        if turn == 0 and piece2 == 16:
-            flag = True
-    if direction == 5:
-        if turn == 1 and piece2 == 2:
-            flag = True
-    if direction == 0 or direction == 3 or direction == 4 or direction == 5:
-        if piece2 == 6 + e_turn * 14 or piece2 == 14 + e_turn * 14:
-            flag = True
-    if direction == 1 or direction == 2 or direction == 6 or direction == 7:
-        if piece2 == 5 + e_turn * 14 or piece2 == 13 + e_turn * 14:
-            flag = True
+    if piece2 == 2 + 14 * e_turn and direction == 5 * turn:
+        flag = True
+    if (piece2 == 6 + e_turn * 14 or piece2 == 14 + e_turn * 14) and (
+        direction == 0 or direction == 3 or direction == 4 or direction == 5
+    ):
+        flag = True
+    if (piece2 == 5 + e_turn * 14 or piece2 == 13 + e_turn * 14) and (
+        direction == 1 or direction == 2 or direction == 6 or direction == 7
+    ):
+        flag = True
     if flag:
         new_array[point1] = _direction_to_pin(direction)
     return new_array
 
 
-def _pin(state: ShogiState):
+# pinされている駒の位置と方向を記録
+def _pin(state: ShogiState) -> np.ndarray:
     bs = _board_status(state)
     turn = state.turn
     pins = np.zeros(81, dtype=np.int32)
@@ -1133,7 +1087,7 @@ def _pin(state: ShogiState):
 
 
 # 特定の方向の動きだけを除いたactionを生成する
-def _eliminate_direction(actions: np.ndarray, direction: int):
+def _eliminate_direction(actions: np.ndarray, direction: int) -> np.ndarray:
     new_array = actions
     # 2方向と成・不成の4方向のフラグを折る
     dir1 = 0
@@ -1168,15 +1122,11 @@ def _eliminate_direction(actions: np.ndarray, direction: int):
 def _kingless_effected_positions(
     bs: np.ndarray, king_point: int, turn: int
 ) -> np.ndarray:
-    all_effect = np.zeros(81)
+    all_effect = np.zeros(81, dtype=np.int32)
     bs[king_point] = 0
     for i in range(81):
-        own = _owner(bs[i])
-        if own != turn:
-            continue
-        piece = bs[i]
-        effect = _piece_moves(bs, piece, i)
-        all_effect += effect
+        if _owner(bs[i]) == turn:
+            all_effect += _piece_moves(bs, bs[i], i)
     return all_effect
 
 
@@ -1199,66 +1149,85 @@ def _king_escape(state: ShogiState) -> bool:
     return flag
 
 
+# point同士の間の位置にフラグを立てる
 def _between(point1: int, point2: int) -> np.ndarray:
-    between = np.zeros(81, dtype=np.int32)
     direction = _point_to_direction(point1, point2, False, 0)
     if direction == -1:
-        return between
-    dif = _direction_to_dif(direction, 0)
-    np.put(between, np.arange(point1 + dif, point2, dif), 1)
-    return between
+        bet = np.zeros(81, dtype=np.int32)
+    else:
+        bet = _change_between(point1, point2, _direction_to_dif(direction, 0))
+    bet[point2] = 0
+    return bet
 
 
+# pinされている駒の非合法手を弾いてlegal_actionを返す
+def _eliminate_pin_actions(
+    bs: np.ndarray, pins: np.ndarray, l_actions: np.ndarray
+):
+    for i in range(81):
+        if pins[i] != 0:
+            l_actions = _filter_action(
+                _eliminate_direction(_create_piece_actions(bs[i], i), pins[i]),
+                l_actions,
+            )
+    return l_actions
+
+
+# 玉が逃げる以外の手に王手回避の手があるかをチェック
+# 存在しない場合True
+# 両王手は考えない(事前にはじく)
+def _is_avoid_check(
+    cn: int,
+    cnp: np.ndarray,
+    cfp: np.ndarray,
+    king_point: int,
+    legal_actions: np.ndarray,
+):
+    # 密接の王手
+    if cn == 1:
+        # 玉が逃げる手以外の合法手は王手をかけた駒がある座標への移動のみ
+        point = int(cnp.argmax())
+        points = cnp
+    # 開き王手
+    else:
+        point = int(cfp.argmax())
+        # pointとking_pointの間。ここに駒を打ったり移動させたりする手は合法
+        points = _between(king_point, point)
+    for i in range(81):
+        if points[i] == 0 and point != i:
+            np.put(legal_actions, np.arange(i, 81 * 34 + i, 81), 0)
+    return (legal_actions == 0).all()
+
+
+# 詰み判定関数
+# 王手がかかっていないならFalse
 def _is_mate(state: ShogiState) -> bool:
     cn, cnp, cf, cfp = _is_check(state)
     # 王手がかかっていないならFalse
     if cn + cf == 0:
         return False
+    else:
+        return _is_mate_noncheck(cn, cnp, cf, cfp, state)
+
+
+# 王手の有無にかかわらず詰みを判定する
+def _is_mate_noncheck(
+    cn: int, cnp: np.ndarray, cf: int, cfp: np.ndarray, state: ShogiState
+):
     legal_actions = _legal_actions(state)
     bs = _board_status(state)
-    # 玉が逃げる手が合法ならその時点で詰んでない
-    if _king_escape(state):
-        return False
     king_point = int(state.board[8 + 14 * state.turn, :].argmax())
+    # 玉が逃げる手以外の合法手
     legal_actions = _filter_move_actions(
         8 + 14 * state.turn, king_point, legal_actions
     )
-    # 玉が逃げる手以外の合法手
-    legal_actions = _filter_action(
-        _create_piece_actions(8 + 14 * state.turn, king_point), legal_actions
-    )
-    # ピンされている駒の動きものぞく
-    pins = _pin(state)
-    for i in range(81):
-        if pins[i] == 0:
-            continue
-        action = _create_piece_actions(bs[i], i)
-        # ピンされた方向以外のaction
-        action = _eliminate_direction(action, pins[i])
-        legal_actions = _filter_action(action, legal_actions)
-    # 2枚以上の駒で王手をかけられた場合、玉を逃げる以外の合法手は存在しない
-    if cn + cf >= 2:
-        return True
-    # 密接の王手
-    if cn == 1:
-        # 玉が逃げる手以外の合法手は王手をかけた駒がある座標への移動のみ
-        point = int(cnp.argmax())
-        for i in range(34):
-            for j in range(81):
-                # 駒打ちは不可
-                if i >= 20:
-                    legal_actions[81 * i + j] = 0
-                else:
-                    # point以外への移動は王手放置
-                    if j != point:
-                        legal_actions[81 * i + j] = 0
-    # 開き王手
-    if cf == 1:
-        point = int(cfp.argmax())
-        # pointとking_pointの間。ここに駒を打ったり移動させたりする手は合法
-        between = _between(king_point, point)
-        for i in range(34):
-            for j in range(81):
-                if between[j] != 1 and not (i <= 19 and j == point):
-                    legal_actions[81 * i + j] = 0
-    return (legal_actions == 0).all()
+    # ピンされている駒の非合法な動きをのぞく
+    legal_actions = _eliminate_pin_actions(bs, _pin(state), legal_actions)
+    # 玉が逃げる手が合法なら詰んでない
+    if _king_escape(state):
+        return False
+    else:
+        # 両王手がかかっている場合はTrue, そうでなければis_avoid_check参照
+        return cn + cf >= 2 or _is_avoid_check(
+            cn, cnp, cfp, king_point, legal_actions
+        )
